package provide Dateentry 0.1
#########################################################################
#                                                                       #
# Script:  dateentry.tcl                                                #
#                                                                       #
# Author:  Dave Freese, dave@darkmatterfactory.com                      #
#                                                                       #
# Provides:  dateentry <widget path> [options]                          #
#                                                                       #
# Purpose:  Provides a mega-widget date entry widget with a dropdown    #
#           calendar provided by calendar.tcl                           #
#           to be chosen.                                               #
#                                                                       #
# Exit Codes:  0  Ok                                                    #
#                                                                       #
# Version:  v0.1.0                                                      #
# --------------------------------------------------------------------- #
#                                                                       #
#                                                                       #
# Version:  0.1.0                                                       #
# Date:    9 October, 2001                                              #
# Author:  Dave Freese, dave@darkmatterfactory.com                      #
# Change:  Original released version.                                   #
#                                                                       #
# Version:                                                              #
# Date:                                                                 #
# Author:                                                               #
# Change:                                                               #
#                                                                       #
#########################################################################

###############################################################################
#    dateentry.tcl - tcl module to provide calendar mega-wdget                #
#    Copyright (C) 2001 Dave Freese                                           #
#                                                                             #
#    This program is free software; you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation; either version 2 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program; if not, write to the Free Software              #
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA  #
#                                                                             #
###############################################################################


package require Tk 8.0
package require Megawidget 0.1
package require Calendar 0.1

# Implements a dateentry widget...

namespace eval ::dateentry {
  namespace export dateentry
  variable widgetOptions
  variable widgetCommands
  variable popupOptions
}

proc ::dateentry::dateentry {w args} {
  variable widgetOptions
  variable widgetCommands

  if {![info exists widgetOptions]} {
    Init
  }
  eval Build $w $args
  return $w
}

proc ::dateentry::Init {} {
  variable widgetOptions
  variable widgetCommands
  variable popupOptions
  
  array set widgetOptions [list                    \
    -background     {background     Background}    \
    -bg             -background                    \
    -foreground     {foreground     Foreground}    \
    -fg             -foreground                    \
    -btnbackground  {btnbackground  BtnBackground} \
    -btnbg          -btnbackground                 \
    -btnforeground  {btnforeground  BtnForeground} \
    -btnfg          -btnforeground                 \
    -dateformat     {dateformat     DateFormat}    \
    -width          {width          Width}         \
    -textvariable   {textVariable   Variable}
  ]

  set widgetCommands [list configure get set]

  image create bitmap ::dateentry::bmpCalendar -data {
    #define bmpCalendar_width 15
    #define bmpCalendar_height 14
    static unsigned char calendar_bits[] = {
      0xaa, 0x37, 0xfc, 0x1a, 0x00, 0x00, 0xff, 0x7f,
      0x55, 0x55, 0xff, 0x7f, 0x55, 0x55, 0xff, 0x7f,
      0x55, 0x55, 0xff, 0x7f, 0x55, 0x55, 0xff, 0x7f,
      0x55, 0x55, 0xff, 0x7f
    }
  }

# The width/height of the popup calendar will be determined when the
# first one is created!  
  set popupOptions(-width) 0
  set popupOptions(-height) 0
}

proc ::dateentry::Build {w args} {
variable widgetOptions
variable widgetCommands
variable popupOptions

  if {[winfo exists $w]} {
    error "window name \"$w\" already exists"
  }

  namespace eval ::dateentry::$w {
    variable widgets
    variable options
    variable txtDate
    
    set widgets(boo) boo
    unset widgets(boo)

    set options(boo) boo
    unset options(boo)
  }

  global ${w}datetemp
  upvar ::dateentry::${w}::widgets widgets
  upvar ::dateentry::${w}::options options
  upvar ::dateentry::${w}::txtDate txtDate
  
# Create the dateentry frame and its subordinates
  set widgets(this)   [frame  $w -class Dateentry -takefocus 0]
  set widgets(entry)  [entry  $widgets(this).e -takefocus 0 \
                         -textvariable txtDate]
  set widgets(btn)    [button $widgets(this).b -takefocus 0 \
                         -image ::dateentry::bmpCalendar \
                         -command "::dateentry::CalPopup $w"]

  set btnbg [$widgets(btn)   cget -background]
  set btnfg [$widgets(btn)   cget -foreground]
  set entbg [$widgets(entry) cget -background]
  set entfg [$widgets(entry) cget -foreground]

  set options(-background)    $entbg
  set options(-foreground)    $entfg
  set options(-btnbackground) $btnbg
  set options(-btnforeground) $btnfg
  set options(-dateformat)    "mm/dd/yyyy"
  set options(-width)         [string length $options(-dateformat)]

  set txtDate ""
  
  set widgets(popup) {}
  set widgets(dateframe) {}
  set widgets(calx) {}
  set widgets(caly) {}
  set widgets(calwidth) 0
  set widgets(calheight) 0

  set widgets(popup) [toplevel ${w}.popup -class CalPopup]
  wm overrideredirect $widgets(popup) 1
  wm withdraw         $widgets(popup)
  set widgets(dateframe) [
    ::calendar::calendar $widgets(popup).date \
      -textvariable ${w}datetemp \
      -dateformat $options(-dateformat) ]
  $widgets(popup) config -cursor left_ptr
  pack $widgets(dateframe)

# set up the keyboard & mouse escape bindings
  set this {[::dateentry::identify %W]}
  bind $widgets(popup) <KeyPress-Escape>   "::dateentry::closepopup $this"
  bind $widgets(popup) <KeyPress-Return>   "::dateentry::closepopup $this"
# cannot bind Any-ButtonPress event to a close or we will miss  the event
# pressing the date button.  the variable is updated on key-press
#  bind $widgets(popup) <Any-ButtonPress>   "::dateentry::closepopup $this"
#  pressing any mouse button outside of the popup calendar will close the
# calendar EXCEPT for a mouse press on the window decoration.
  bind $widgets(popup) <Any-ButtonRelease> "::dateentry::checkbtn $this %x %y %X %Y"
  bind $widgets(popup) <FocusOut>      "::dateentry::closepopup $this"
  
# Retrieve options from .Xdefaults file & set the values,

  foreach name [array names widgetOptions] {
    if {[llength $widgetOptions($name)] == 1} continue
    set optName  [lindex $widgetOptions($name) 0]
    set optClass [lindex $widgetOptions($name) 1]
    set value [option get $w $optName $optClass]
    if {[string length $value] != 0} {
      set options($name) $value
    }
  }

# Now take care of configuration items passed on the arguement list...

  if { [llength $args] > 0 } {
    if { [catch "::dateentry::Configure $widgets(this) $args" error] } {
      catch {destroy $w}
      error $error
    }
  }

# Store the proc to control the widget

  set widgets(frame) ::calendar::${w}::${w}

  $widgets(entry) configure -width $options(-width)
  
  pack $widgets(entry)  -side left -expand yes -fill x
  pack $widgets(btn)    -side left -expand no

  rename ::$w $widgets(frame)

  bind dateentry <Destroy> [list ::megawidget::DestroyHandler %W]

  proc ::$w {command args} \
    "eval ::dateentry::WidgetProc $w \$command \$args"

  return ""
}

proc ::dateentry::Configure {w args} {
  variable widgetOptions
  upvar ::dateentry::${w}::widgets widgets
  upvar ::dateentry::${w}::options options

  # Return all results to the user
  if {[llength $args] == 0} {
    set results {}
    foreach opt [lsort [array names widgetOptions]] {
      if {[llength $widgetOptions($opt)] == 1} {
        set alias $widgetOptions($opt)
        set optName $widgetOptions($alias)
        lappend results [list $opt $optName]
      } else {
        set optName  [lindex $widgetOptions($opt) 0]
        set optClass [lindex $widgetOptions($opt) 1]
        set default [option get $w $optName $optClass]
        set setopt {}
        if {[info exists options($opt)]} {set setopt $options($opt)}
        lappend results [list $opt $optName $optClass $default $setopt]
      }
    }
    return $results
  }

  # Are we looking for config info for an option?
  if {[llength $args] == 1} {
    set opt [::megawidget::Canonize $w option $args]
    set optName  [lindex $widgetOptions($opt) 0]
    set optClass [lindex $widgetOptions($opt) 1]
    set default [option get $w $optName $optClass]
    set setopt {}
    if {[info exists options($opt)]} {set setopt $options($opt)}
    if {![info exists options($opt)]} {
      set found [array get options]
      puts "$opt not found in:\n $found"
      }
    set results [list $opt $optName $optClass \
      $default $setopt]
    return $results
  }

  # if we have an odd number of values, bail.
  if {[expr {[llength $args]%2}] == 1} {
    # hmmm. An odd number of elements in args
    error "value for \"[lindex $args end]\" missing"
  }

  # deal with arg/value pairs, putting them in an array...
  foreach {name value} $args {
    set name [::megawidget::Canonize $w option $name]
    set opts($name) $value
  }

  foreach option [array names opts] {
    set newValue $opts($option)
    if {[info exists options($option)]} {
      set oldValue $options($option)
    }

    switch -- $option {
      -background {
        if { "$newValue" != "" } {
          $widgets(entry) configure -background $newValue
        }
        set options($option) $newValue
      }
      -foreground {
        if { "$newValue" != "" } {
          $widgets(entry) configure -foreground $newValue
          set options($option) $newValue
        }
      }
      -btnbackground {
        if { "$newValue" != "" } {
          $widgets(btn) configure -background $newValue
        }
        set options(btn) $newValue
      }
      -btnforeground {
        if { "$newValue" != "" } {
          $widgets(btn) configure -foreground $newValue
          set options($option) $newValue
        }
      }
      -width {
        if { "$newValue" != "" } {
          $widgets(entry) configure -width $newValue
        }
        set options($option) $newValue
      }
      -dateformat {
        if {"$newValue" != "" } {
          set options($option) $newValue
          $widgets(entry) config -width [string length $options(-dateformat)]
        }
      }
      -textvariable {
        $widgets(entry) configure -textvariable $newValue
        set options($option) $newValue
      }
    }
  }
}

# This is the actual command that gets called when the user attempts
# to use the widget object name as a procedure.

proc ::dateentry::WidgetProc {w command args} {
upvar ::dateentry::${w}::widgets widgets
upvar ::dateentry::${w}::options options
upvar ::dateentry::${w}::txtDate txtDate

  set command [::megawidget::Canonize $w command $command]

  set result ""

  switch $command {
    configure {
      set result [eval ::dateentry::Configure {$w} $args]
    }
    set {
      return ""
    }
    get {
      return [$txtDate]
    }
    default {
      error "bad option \"$command\""
    }
  }
}

proc ::dateentry::CalPopup {w} {
upvar ::dateentry::${w}::widgets widgets
upvar ::dateentry::${w}::options options
upvar ::dateentry::${w}::txtDate txtDate
global ${w}datetemp
upvar #0 ${w}datetemp temp

# determine location of lower left corner of parent widget
# window may have been moved by user, so must compute each time
# the calendar popup is invoked
  set  widgets(caly) [winfo rooty $widgets(entry)]
  incr widgets(caly) [winfo height $widgets(entry)]
  set  widgets(calx) [winfo rootx $widgets(entry)]

  if {$widgets(calwidth) == 0} {
    set widgets(calwidth)  [winfo reqwidth $widgets(dateframe)]
    set widgets(calheight) [winfo reqheight $widgets(dateframe)]
  }

  wm geometry $widgets(popup) \
    $widgets(calwidth)\x$widgets(calheight)+$widgets(calx)+$widgets(caly)
  $widgets(dateframe) configure \
    -dateformat $options(-dateformat) \
    -cdate [$widgets(entry) get]
  

  wm deiconify $widgets(popup) 
  raise $widgets(popup)

  set temp ""
  set widgets(focus) [focus]
  focus $widgets(popup)

# need to use the global grab to insure that any click outside of the
# popup calendar will close the popup!  if the local grab is used then
# a click on the title bar of any other window including the parent window
# will not close the popup.
  
  grab -global $widgets(popup)
  tkwait variable ${w}datetemp
  grab release $widgets(popup)
  focus $widgets(focus)
  wm withdraw $widgets(popup)

  if { $temp == "esc" } return

  set oldstate [$widgets(entry) cget -state]
  $widgets(entry) configure -state normal
  $widgets(entry) delete 0 end
  $widgets(entry) insert 0 $temp
  $widgets(entry) configure -state $oldstate
}

proc ::dateentry::closepopup {w} {
upvar #0 ${w}datetemp temp
  set temp "esc"
}

proc ::dateentry::checkbtn {w x y X Y} {
upvar #0 ${w}datetemp temp
upvar ::dateentry::${w}::widgets widgets
  if {$x < 0 || 
      $y < 0 || 
      $x > $widgets(calwidth) || 
      $y > $widgets(calheight)} {
    set temp "esc"
  }
}

proc ::dateentry::identify {w} {
  set result {}
  if {![winfo exists $w]} {
    error "window \"$w\" doesn't exist"
  }
  set win $w
  while {[winfo class $win] != "CalPopup"} {
    set win [winfo parent $win]
    if {$win == "."} break;
  }
  if {[winfo class $win] == "CalPopup"} {
    set win [winfo parent $win]
  }
  lappend result $win
  return $result
}
